/*
 * manager-routes.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2025 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <linux/if.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-manager.h"

/* Empaquetador de datos */
int _manager_pack_route_table_info (unsigned char *buffer, uint32_t table_index, const char *table_name) {
	int name_len;
	
	memcpy (&buffer[0], &table_index, 4);
	
	name_len = strlen (table_name);
	buffer[4] = name_len;
	memcpy (&buffer[5], table_name, name_len);
	
	return 5 + name_len;
}

/* Los comandos que verdaderamente envian */
void _manager_send_route_table (ManagerClientInfo *manager_client, uint32_t table_index, const char *table_name, uint8_t is_event) {
	unsigned char buffer[300];
	int size;
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_ROUTE_TABLE_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_ROUTE_TABLE;
	}
	
	size = 2 + _manager_pack_route_table_info (&buffer[2], table_index, table_name);
	
	send (manager_client->fd, buffer, size, 0);
}

void _manager_send_route_table_deleted (ManagerClientInfo *manager_client, uint32_t table_index) {
	unsigned char buffer[80];
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_ROUTE_TABLE_REMOVED;
	
	memcpy (&buffer[2], &table_index, 4);
	
	send (manager_client->fd, buffer, 6, 0);
}

void _manager_route_table_handle_list_tables (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int g, table_count;
	uint32_t table, table_wanted;
	char table_name[256];
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_LIST_ROUTE_TABLES);
		return;
	}
	
	memcpy (&table_wanted, &buffer[2], 4);
	
	table_count = network_inador_rtables_get_count_tables (manager_client->manager->handle);
	for (g = 0; g < table_count; g++) {
		network_inador_rtables_get_table_by_index (manager_client->manager->handle, g, &table, table_name, sizeof (table_name));
		
		if (table_wanted != 0 && table_wanted != table) continue;
		
		_manager_send_route_table (manager_client, table, table_name, FALSE);
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_ROUTE_TABLES);
}

/* Los eventos que vienen desde la librería */
void _manager_route_table_added_event_cb (NetworkInadorHandle *handle, uint32_t table_index, const char *table_name, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando tabla de ruteo agregada: %s (%i)\n", table_name, table_index);
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_ROUTE_TABLES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_route_table (manager_client, table_index, table_name, TRUE);
		}
	}
}

void _manager_route_table_deleted_event_cb (NetworkInadorHandle *handle, uint32_t table_index, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando tabla de ruteo eliminada: %i\n", table_index);
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_ROUTE_TABLES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_route_table_deleted (manager_client, table_index);
		}
	}
}

